'''
Created on Feb 14, 2013
@author: Luke Allen
To make this program work, you need to also install:
-NLTK and its dependencies by following http://nltk.org/install.html
-JPype from http://jpype.sourceforge.net/
-Download the Stanford Parser from http://nlp.stanford.edu/software/lex-parser.shtml
 and copy its .jar files to a folder called "../jar/" relative to this .py file, 
 or else change the line labeled "directory that .jar files are in" below.
'''
#import nltk
from nltk.corpus import wordnet as wn
from nltk.corpus import verbnet as vn
import poetry
from _pyio import open
from nltk import tokenize
import datetime
import jpype
import os.path
jarpath = os.path.join(os.path.abspath('.'), "../jar") #directory that .jar files are in, relative to this .py script.
jpype.startJVM(jpype.getDefaultJVMPath(),"-ea", "-Djava.ext.dirs=%s" % jarpath) #start a JVM
nlp = jpype.JPackage("edu.stanford.nlp") #this is how you import a package (which contains classes). Now we can use the classes inside it, almost as easily as Java.
#LexicalizedParser = jpype.JClass("edu.stanford.nlp.parser.lexparser.LexicalizedParser")  # can also import a specific Java class like this. Repeat this for any more classes you need.
lp = nlp.parser.lexparser.LexicalizedParser.loadModel() #call a static method of a class like this (see the Stanford Parser's Javadoc for a description of its methods and classes) 
tlp = nlp.trees.PennTreebankLanguagePack() #now tlp is an instance of TreebankLanguagePack
gsf = tlp.grammaticalStructureFactory()




#lp = LexicalizedParser.loadModel("edu/stanford/nlp/models/lexparser/englishPCFG.ser.gz") #call a static method of the class. Doesn't seem to work.


#takes a wordnet lemma, whose string representation might look like "Lemma('car.n.01.automobile')" and returns "automobile"
def makeWordnetLemmaReadable(lemma):
    #print lemma
    parts = str(lemma).rsplit(".",1)#split into a list, starting on the right, separated by ".", limit of 1 split
    s = parts[-1]#take last part
    s = s[:-2].replace("_", " ")#remove the ending "')" from the string, and also replace underscores with spaces
    #print s
    return s

#Construct this by passing in a word. It builds an internal list of its synonyms and the pronunciation of each.
class RhymeToken:
    def __init__(self, word, index, POS = None):
        self.index = index # my index in the global tokenList
        self.mySentence = None # will be filled in later
        self.isStop = False #True if this is a special empty token I put at sentence breaks.
        self.isStart = False #Was first word in a sentence originally.
        self.dListsImIn = []
        self.rhymeScoreCache = dict() #keys are other tokens. Values are the highest rhyme score with that token, or 0 if there was no rhyme.
        self.phraseEnd = False #gets set later
        if(not isPunctuation(word)):
            self.makesSound = True
            self.POS = POS
            hypernymLevelsToExplore = 3
            s = Synonym(word, 2.0) #The Synonym object for the original word. Give it a score above 1 to ensure it's the highest-scoring Synonym for this token
            self.synonyms = addPhonemesGivenSynonymObjects(  getSynonyms(word, POS, hypernymLevelsToExplore,  [s])  )
            self.tokensIRhymeWith = []#gets filled in later
#            if len(self.synonyms) <=1: #uncomment to flag words for debugging
#                self.synonyms[0].word = ":("+self.synonyms[0].word
        else: # is punctuation
            self.makesSound = False
            if word == "\n":
                self.isStop = True
                self.synonyms = [Synonym("",1)]
            else:
                self.synonyms = [Synonym(word,1)]
                
    def __repr__(self):
        return self.synonyms[0].word
    
class Synonym:
    def __init__(self, word, score):
        self.word = word
        self.similarityScore = score
        self.phonemes = []
        self._nSyllablesCached = -1 #a value of -1 signifies that we haven't calculated it. (Can't use "False" because 0 also evaluates to False)
    def addPhonemes(self, phonemes):
        self.phonemes = tuple(phonemes)#make it a fixed-size list for speed
    def numSyllables(self):#get number of syllables of this word (I lazily calculate this because it's often never needed)
        if self._nSyllablesCached < 0:
            self._nSyllablesCached = poetry.nsylHandlingSpaces(self.word)
        return self._nSyllablesCached
    def __repr__(self):
        return self.word


#Returns false if the word is actually punctuation that shouldn't be rhymed.
def isPunctuation(word):
    return not word.isalnum()

def stanfordPOSToWordnetPOS(POS):
    #converts the part-of-speech labels of the Stanford Parser into those used by WordNet. Returns false if WordNet doesn't have that part of speech.
    if not POS:
        return False
    elif POS[0]=="N":
        return wn.NOUN
    elif POS[0]=="V":
        return wn.VERB
    elif POS[0]=="R":
        return wn.ADV
    elif POS[0]=="J":
        return wn.ADJ
    return False

WORDS_WHERE_WORDNET_FAILS = {"have", "had", "has", "having", "was", "is", "were", "be", "am", "are" }#don't try to get synonyms for these words. WordNet doesn't include the most common stopword use of "have".
        
#TODO: pass in part of speech and constrain WordNet with it
#TODO: sources of synonyms besides wordnet
#TODO: use antonyms, and prepend "not" for adjectives and "did/does not" for verbs
def getSynonyms(word, POS, hypernymLevelsToExplore, synonymsListToAddTo):
    wordNetPOS = stanfordPOSToWordnetPOS(POS)
    if not wordNetPOS: #it's a stop word that isn't in WordNet
        return synonymsListToAddTo
    #TODO: maybe lemmatize
    if word.lower() in WORDS_WHERE_WORDNET_FAILS:#don't try to get synonyms for these words
        return synonymsListToAddTo
    allSynsets = wn.synsets(word, wordNetPOS)
    originalWordLemma = wn.morphy(word, wordNetPOS)#lemmatized form of the original word. E.g. actions -> action
    #synonyms
    senseLikelihood = .95
    if wordNetPOS == wn.VERB:#add verbnet synonym list
        for classid in vn.classids(originalWordLemma):
            for lemma in vn.lemmas(classid):
                synonymsListToAddTo.append( Synonym(lemma, senseLikelihood ))
    senseLikelihood = 1.0
    for i, synset in enumerate(allSynsets):
        if i>0:#Using most-common synset instead of doing WSD. TODO: implement WSD 
            #break #use this line to only use the most common synset
            senseLikelihood = .6 #use this line to use all synsets
        #print synset.definition
        for lemma in synset.lemmas: #add synonyms
            readableLemma = makeWordnetLemmaReadable(lemma)
            if readableLemma != originalWordLemma: #the original word is part of the synset. Don't put it into the synonym list; we already have it!
                synonymsListToAddTo.append( Synonym(readableLemma, senseLikelihood ))
#            else:
#                print "ignoring duplicate lemma for "+originalWordLemma
        if wordNetPOS != wn.VERB:#Don't do hyper/hyponyms for verbs because those seem to have hugely different meanings in WordNet
            senseLikelihood *= .7#lower score for hyper/hyponyms
            if synset.hypernyms():
                hypernymSynset = synset.hypernyms()[0]#This seems to return a list that's always size 1.
                while(hypernymLevelsToExplore > 0):
                    for lemma in hypernymSynset.lemmas: #add synonyms
                        synonymsListToAddTo.append( Synonym(makeWordnetLemmaReadable(lemma), senseLikelihood ))
                    hypernymLevelsToExplore-=1
                    if not hypernymSynset.hypernyms():
                        break
                    hypernymSynset = hypernymSynset.hypernyms()[0]
            #add any hyponyms. These have a high rate of changing the sentence meaning.
#            for relatedSynset in synset.hyponyms():
#                for lemma in relatedSynset.lemmas: #add synonyms
#                    synonymsListToAddTo.append( Synonym(makeWordnetLemmaReadable(lemma), senseLikelihood ))
        #TODO: add "not" + lemma.antonyms() 
    return synonymsListToAddTo #todo: score each with its appropriateness. Either with network distance or context.

#Pass in a list of Synonym objects. Fills in their phonemes, 
#and returns a list containing the first one (always) and all subsequent ones (if we could find their phonemes). 
def addPhonemesGivenSynonymObjects(synonymList):
    if synonymList:
        assert isinstance(synonymList[0], Synonym)
    pairs = []
    for s in synonymList:
        phonemes = poetry.phonemes(s.word)
        if phonemes or s is synonymList[0]:#only keep the word in the synonyms list if we actually know its pronunciation, or if it's the original word
            s.addPhonemes(phonemes)
            pairs.append( s )
    return pairs

#OLD VERSION: Pass in a list of words. Returns a list of tuples of [word, [CMU dictionary phonemes]]
def addPhonemesGivenWords(synonymList):
    #replace each word with a [word, phonemes] tuple.
    pairs = []
    for i in range(len(synonymList)):
        phonemes = poetry.phonemes(synonymList[i])
        if phonemes:#only keep the word in the list if we actually know its pronunciation
            pairs.append([synonymList[i], phonemes])
    return pairs


def bestRhymes(a, b):#Pass in two RhymeTokens. Returns a sorted list of tuples containing (rhyme score, synonym from a, synonym from b)
    assert isinstance(a, RhymeToken)
    assert isinstance(b, RhymeToken)
    rhymePairs = []
    for synA in a.synonyms:
        for synB in b.synonyms:
            if(synA.word.rstrip()[-1] != synB.word.rstrip()[-1]): #check that we're not trying to rhyme a word to itself (the rstrip makes us just check the last word)
                score = poetry.rhyme_score_from_phonemes(synA.phonemes, synB.phonemes)
                if(score >= 1):
                    rhymePairs.append( ( score*synA.similarityScore*synB.similarityScore, synA, synB ) )
    if rhymePairs:#is not empty
        rhymePairs = sorted(rhymePairs, key=lambda element: -element[0]) #sort, highest first
        return rhymePairs  #[0:max(len(rhymePairs)-1,10)]
    else:
        return rhymePairs




#grabs a range of word tokens for a rhyme check, skipping punctuation.
def wordsInSequence(tokenList, start, howFar):
    sequence = []
    wordIndex=start-1
    while howFar > 0:
        wordIndex+=1
        if wordIndex >= len(tokenList):
            return sequence
        if tokenList[wordIndex].makesSound:
            howFar-=1#TODO: subtract syllable count
            tokenList[wordIndex].tokensIRhymeWith=[]
            sequence.append(tokenList[wordIndex])
    return sequence
    

def findPossibleRhymes(tokenList, startIndex, numSyllablesToSearch):#looks over the given search space and returns a best-first list of the
                                        #words that can be made to rhyme, in the form of tuples of (rhymeScore, tokenA, tokenB).
    assert isinstance(tokenList[0], RhymeToken)
    rhymingTokenPairs = []
    syllablesCoveredForI = 0 #how many syllables the word is away from startIndex
    for i in range(startIndex, len(tokenList)):#we'll also limit the loop by number of syllables, below
        tokenI = tokenList[i]
        if tokenI.phraseEnd:
            tokenIBonus = BONUS_FACTOR_FOR_RHYME_ON_PHRASE_END
        else:
            tokenIBonus = 1.0
        assert isinstance(tokenI.synonyms[0], Synonym)
        syllablesCoveredForI += tokenI.synonyms[0].numSyllables()
        if syllablesCoveredForI > numSyllablesToSearch:
            break#past our syllable limit; leave the loop
        syllablesCoveredForJ = 0
        for j in range(i+1, len(tokenList)):#we'll also limit the loop by number of syllables, below
            tokenJ = tokenList[j]
            if tokenJ.phraseEnd:
                tokenJBonus = BONUS_FACTOR_FOR_RHYME_ON_PHRASE_END
            else:
                tokenJBonus = 1.0
            syllablesCoveredForJ += tokenJ.synonyms[0].numSyllables()
            if syllablesCoveredForI + syllablesCoveredForJ > numSyllablesToSearch:
                break#past our syllable limit; leave the loop
            #if here, we have a valid i and j in the range
            if tokenI.makesSound and tokenJ.makesSound:
                if tokenJ in tokenI.rhymeScoreCache:#we've compared these tokens before; use the cached score of the best rhyme we found
                    cachedScore = tokenI.rhymeScoreCache[tokenJ]
                    if cachedScore > 0:
                        rhymingTokenPairs.append(  (cachedScore, tokenI, tokenJ )  )
                else:#we haven't compared these tokens before
                    rhymesForTheseTokens = bestRhymes(tokenList[i], tokenList[j])
                    if(rhymesForTheseTokens):
                        topRhymeScore = rhymesForTheseTokens[0][0]*tokenIBonus*tokenJBonus
                        rhymingTokenPairs.append(  (topRhymeScore, tokenList[i], tokenList[j] )  )
                        tokenI.rhymeScoreCache[tokenJ] = topRhymeScore
                        tokenJ.rhymeScoreCache[tokenI] = topRhymeScore
                    else:
                        tokenI.rhymeScoreCache[tokenJ] = 0
                        tokenJ.rhymeScoreCache[tokenI] = 0
    return sorted(rhymingTokenPairs, key=lambda element: -element[0]) #sort, highest first



def printBestSingleRhymeForTesting(tokenList):#adds line breaks at the best rhymes, with no attempt to match line syllable counts
    startingTokenIndex=0
    while startingTokenIndex < len(tokenList):
        bestRhymePairs = findPossibleRhymes(tokenList, startingTokenIndex, MAX_SYLLABLES_TO_SEARCH_OVER)
        if not bestRhymePairs:
            print str(tokenList[startingTokenIndex])
            startingTokenIndex += 1
        else:
            bestTokenA = bestRhymePairs[0][1]
            bestTokenB = bestRhymePairs[0][2]
            bestRhymePair = bestRhymes(bestTokenA,bestTokenB)[0]
            line = ""
            for wordIndex in range(startingTokenIndex, bestTokenA.index):
                line += " " + str(tokenList[wordIndex])
            line += " " + str(bestRhymePair[1])#print rhyming synonym instead of word
            print line
            line = ""
            for wordIndex in range(bestTokenA.index+1, bestTokenB.index):
                line += " " + str(tokenList[wordIndex])
            line += " " + str(bestRhymePair[2]) + "                              **from "+str(len(bestRhymePairs))+" possible rhymes" #print rhyming synonym instead of word
            print line
            startingTokenIndex = bestTokenB.index+1
            
def printBestSingleRhymeEqualized(tokenList):#adds line breaks at the best rhymes, with no attempt to match line syllable counts
    startingTokenIndex=0
    while startingTokenIndex < len(tokenList):
        bestRhymePairs = findPossibleRhymes(tokenList, startingTokenIndex, MAX_SYLLABLES_TO_SEARCH_OVER)
        if not bestRhymePairs:
            print str(tokenList[startingTokenIndex])
            startingTokenIndex += 1
        else:
            bestTokenA = bestRhymePairs[0][1]
            bestTokenB = bestRhymePairs[0][2]
            bestRhymePair = bestRhymes(bestTokenA,bestTokenB)[0]
            lineA = tokenList[startingTokenIndex : bestTokenA.index]#notation means start:(one past end)
            lineB = tokenList[bestTokenA.index+1 : bestTokenB.index]
            print "trying to equalize these two: "
            print lineA
            print lineB
            aSyl = bestRhymePair[1].numSyllables()
            bSyl = bestRhymePair[2].numSyllables()
            (_deletionPenalty, equalizedLines) = equalizeLines(lineA, lineB, aSyl-bSyl, MIN_SYLLABLES_PER_LINE-aSyl, MAX_SYLLABLES_PER_LINE-aSyl)
            if not equalizedLines: #couldn't equalize for this rhyme pair
                print "****equalize failed:****" 
                line = ""
                for wordIndex in range(startingTokenIndex, bestTokenA.index-1):
                    line += " " + str(tokenList[wordIndex])
                line += " " + str(bestRhymePair[1])#print rhyming synonym instead of word
                print line
                line = ""
                for wordIndex in range(bestTokenA.index+1, bestTokenB.index-1):
                    line += " " + str(tokenList[wordIndex])
                line += " " + str(bestRhymePair[2]) + "                              **from "+str(len(bestRhymePairs))+" possible rhymes" #print rhyming synonym instead of word
                print line
            else:
                print "**** equalized: *****"
                line = ""
                for token in equalizedLines[0]:
                    line += " " + str(token)
                line += " " + str(bestRhymePair[1])#print rhyming synonym instead of word
                print line
                line = ""
                for token in equalizedLines[1]:
                    line += " " + str(token)
                line += " " + str(bestRhymePair[2]) + "                              **from "+str(len(bestRhymePairs))+" possible rhymes" #print rhyming synonym instead of word
                print line
            startingTokenIndex = bestTokenB.index+1


def printSearchedRhymesEqualized(tokenList):#adds line breaks at the best rhymes, with no attempt to match line syllable counts
    startingTokenIndex=0
    MAX_RHYMES_TO_SEARCH = 1000 #turns out the search is fast, so no need to limit it.
    while startingTokenIndex < len(tokenList):
        bestRhymePairs = findPossibleRhymes(tokenList, startingTokenIndex, MAX_SYLLABLES_TO_SEARCH_OVER)
        if not bestRhymePairs:
            print str(tokenList[startingTokenIndex])
            startingTokenIndex += 1
        else:
            foundSolution = False
            for rhymeIndex in range(0, min(MAX_RHYMES_TO_SEARCH, len(bestRhymePairs))):
                #we have a bunch of rhymes; search the highest-scoring first
                bestTokenA = bestRhymePairs[rhymeIndex][1]
                bestTokenB = bestRhymePairs[rhymeIndex][2]
                bestRhymePair = bestRhymes(bestTokenA,bestTokenB)[0]
                lineA = tokenList[startingTokenIndex : bestTokenA.index]#notation has the form [start:(one past end)]
                lineB = tokenList[bestTokenA.index+1 : bestTokenB.index]
                aSyl = bestRhymePair[1].numSyllables()
                bSyl = bestRhymePair[2].numSyllables()
                (_deletionPenalty, equalizedLines) = equalizeLines(lineA, lineB, aSyl-bSyl, MIN_SYLLABLES_PER_LINE-aSyl, MAX_SYLLABLES_PER_LINE-aSyl, IDEAL_SYLLABLES_PER_LINE-aSyl)
                if equalizedLines:
                    line = ""
                    for token in equalizedLines[0]:
                        line += " " + str(token)
                    line += " " + str(bestRhymePair[1])#print rhyming synonym instead of word
                    print line
                    line = ""
                    for token in equalizedLines[1]:
                        line += " " + str(token)
                    line += " " + str(bestRhymePair[2]) #+ "                              ****** equalized: ***** from "+str(len(bestRhymePairs))+" possible rhymes" #print rhyming synonym instead of word
                    print line
                    
                    foundSolution = True
                    break #TODO: keep track of best solution, including paraphrase cost, and print it at the end
            if not foundSolution: #just print the last rhyme pair we looked at and move on
                line = ""
                for wordIndex in range(startingTokenIndex, bestTokenA.index-1):
                    line += " " + str(tokenList[wordIndex])
                line += " " + str(bestRhymePair[1])#print rhyming synonym instead of word
                print line
                line = ""
                for wordIndex in range(bestTokenA.index+1, bestTokenB.index-1):
                    line += " " + str(tokenList[wordIndex])
                line += " " + str(bestRhymePair[2]) + "                              ******equalize failed:**** from "+str(len(bestRhymePairs))+" possible rhymes" #print rhyming synonym instead of word
                print line
            startingTokenIndex = bestTokenB.index+1

def printSearchedRhymesRearranged_firstSolution(tokenList):#adds line breaks at the best rhymes, with no attempt to match line syllable counts
    startingTokenIndex=0
    MAX_RHYMES_TO_SEARCH = 1000 #turns out the search is fast, so no need to limit it.
    overflow = ""
    while startingTokenIndex < len(tokenList):
        if not tokenList[startingTokenIndex].makesSound: #skip any punctuations left over from the last line.
            startingTokenIndex+=1
            continue#back to outer while loop
        bestRhymePairs = findPossibleRhymes(tokenList, startingTokenIndex, MAX_SYLLABLES_TO_SEARCH_OVER)
        if not bestRhymePairs:#no rhyme pairs found in the sample. Usually only happens for the last couple words
            overflow += " " + str(tokenList[startingTokenIndex])
            startingTokenIndex += 1
        else:
            if overflow:
                print overflow + "             *****  <--no rhymes"
            foundSolution = False
            for rhymeIndex in range(0, min(MAX_RHYMES_TO_SEARCH, len(bestRhymePairs))):
                #we have a bunch of rhymes; search the highest-scoring first
                bestTokenA = bestRhymePairs[rhymeIndex][1]
                bestTokenB = bestRhymePairs[rhymeIndex][2]
                bestRhymePair = bestRhymes(bestTokenA,bestTokenB)[0]
                endIndex = tokenList.index(bestTokenB)+1
                while(not tokenList[endIndex].isStop):#find last token that might be involved in within-sentence movements. I.e. the end of the second token's sentence.
                    endIndex += 1
                window = tokenList[startingTokenIndex : endIndex+1]
                aSyl = bestRhymePair[1].numSyllables()
                bSyl = bestRhymePair[2].numSyllables()
                relevantMovements = getRelevantRearrangements(startingTokenIndex, tokenList.index(bestTokenA), tokenList.index(bestTokenB))
                possibleLines = []
                swapAAndB = False
                n = len(relevantMovements)
                if n <= 14:#try all combinations of movements. This is O(2^n) so we need to limit n
                    maxCombo = 2 ** n
                    for combo in range(maxCombo):#for each possible True/False sequence, apply that combination of movements. Note that the first will be all 0's, which covers the do-nothing option
                        perturbedWindow = list(window)
                        listOfMovements = []
                        movementPenalty = 1.0
                        for index, movement in enumerate(relevantMovements):
                            if ((combo >> index)&0x01) > 0:
                                listOfMovements.append(movement)
                                movement.applyToTokenList(perturbedWindow, False)#apply the movement and get a new, post-movement sequence of tokens
                                movementPenalty *= movement.penaltyFactor
                        score_lines_swap = getEqualizedLinesGivenWindow(perturbedWindow, bestTokenA, bestTokenB, aSyl, bSyl)
                        if score_lines_swap:
                            score_lines_swap[0] = score_lines_swap[0] * movementPenalty
                            score_lines_swap.append(listOfMovements)
                            score_lines_swap_movements = score_lines_swap #just documenting what it holds
                            possibleLines.append(score_lines_swap_movements)
                else:#if n is too large, just try to apply each movement individually
                    relevantMovements.append(MovementList(bestTokenA,1.0))#include a do-nothing movement
                    for mList in relevantMovements:#try applying each movement individually
                        perturbedWindow = list(window)
                        mList.applyToTokenList(perturbedWindow, False)#apply the movement and get a new, post-movement sequence of tokens
                        score_lines_swap = getEqualizedLinesGivenWindow(perturbedWindow, bestTokenA, bestTokenB, aSyl, bSyl)
                        if score_lines_swap:
                            score_lines_swap[0] = score_lines_swap[0] * mList.penaltyFactor
                            score_lines_swap.append([mList])
                            score_lines_swap_movements = score_lines_swap #documenting what it holds
                            possibleLines.append(score_lines_swap_movements)
                    
                if possibleLines: #found at least one solution for this rhyme. Output the best one and exit.
                    foundSolution = True
                    possibleLines.sort()#defaults to sorting by first thing in the tuple
                    (_score, equalizedLines, swapAAndB, mLists) = possibleLines[-1]#highest score
                    Aword = bestRhymePair[1]
                    Bword = bestRhymePair[2]
                    if swapAAndB:
                        Aword,Bword = Bword,Aword
                        bestTokenA, bestTokenB = bestTokenB, bestTokenA
                    for mList in mLists:
                        mList.applyToTokenList(tokenList, True)#do the movement to the global list
                    #print the lines
                    printPoetryLine(equalizedLines[0], bestTokenA, Aword)
                    printPoetryLine(equalizedLines[1], bestTokenB, Bword)
                    startingTokenIndex = tokenList.index(bestTokenB)+1#+skipInNextLine
                    break #Stop the search. TODO: keep track of best solution across all rhymes, including paraphrase cost, and print it at the end
            if not foundSolution: #just print the last rhyme pair we looked at and move on
                bestTokenA = bestRhymePairs[0][1]
                bestTokenB = bestRhymePairs[0][2]
                printPoetryLine(tokenList[startingTokenIndex:tokenList.index(bestTokenA)], bestTokenA, bestRhymePair[1])
                printPoetryLine(tokenList[tokenList.index(bestTokenA)+1:tokenList.index(bestTokenB)], bestTokenB, bestRhymePair[2])
                startingTokenIndex = tokenList.index(bestTokenB)+1

def printSearchedRhymesRearranged(tokenList):#adds line breaks at the best rhymes, with no attempt to match line syllable counts
    startingTokenIndex=0
    overflow = ""
    while startingTokenIndex < len(tokenList):
        if not tokenList[startingTokenIndex].makesSound: #skip any punctuations left over from the last line.
            startingTokenIndex+=1
            continue#back to outer while loop
        bestRhymePairs = findPossibleRhymes(tokenList, startingTokenIndex, MAX_SYLLABLES_TO_SEARCH_OVER)
        if not bestRhymePairs:#no rhyme pairs found in the sample. Usually only happens for the last couple words. #Add the current word and any punctuation after it to the print queue.
            overflow += " " + str(tokenList[startingTokenIndex])
            startingTokenIndex += 1
            while startingTokenIndex < len(tokenList) and not tokenList[startingTokenIndex].makesSound:
                overflow += str(tokenList[startingTokenIndex])
                startingTokenIndex += 1
            continue
        bestScoreIncludingPenalty = 0
        possibleLines = []
        for rhymingTokens in bestRhymePairs:#we have a bunch of rhymes (sorted best first); search them for possible line-length matches
            baseRhymeScoreWithNoPenalty = rhymingTokens[0]
            if baseRhymeScoreWithNoPenalty < bestScoreIncludingPenalty:#can't possibly beat our best score. Because they're sorted by baseRhymeScore, we can end the search if we see this. 
                break
            score_lines_swap_movements = getBestPossibleLines(rhymingTokens, startingTokenIndex)
            if score_lines_swap_movements:#at least one solution; add it.
                score_lines_swap_movements[0] *= baseRhymeScoreWithNoPenalty
                bestScoreIncludingPenalty = max(bestScoreIncludingPenalty, score_lines_swap_movements[0])
                score_lines_swap_movements.append(rhymingTokens)
                possibleLines.append(score_lines_swap_movements)
        if possibleLines: #found at least one solution. Output the best one and exit.
            if overflow:#flush the buffer of failed lines
                print overflow + "             *****  <--no rhymes"
                overflow = ""
            possibleLines.sort()#defaults to sorting by first thing in the tuple
            (_score, equalizedLines, swapAAndB, mLists, rhymingTokens) = possibleLines[-1]#highest score
            if swapAAndB:
                bestTokenA, bestTokenB = rhymingTokens[2], rhymingTokens[1]#index 0 holds their score
            else:
                bestTokenA, bestTokenB = rhymingTokens[1], rhymingTokens[2]
            bestRhymePair = bestRhymes(bestTokenA,bestTokenB)[0]
            Aword = bestRhymePair[1]
            Bword = bestRhymePair[2]
            for mList in mLists:
                mList.applyToTokenList(tokenList, True)#do the movement to the global list
#                print "executing "+str(mList)
            #print the lines
            printPoetryLine(equalizedLines[0], bestTokenA, Aword)
            printPoetryLine(equalizedLines[1], bestTokenB, Bword)
            startingTokenIndex = tokenList.index(bestTokenB)+1#+skipInNextLine
        else: #No solutions. Just move on. Add the current word and any punctuation after it to the print queue.
            overflow += " " + str(tokenList[startingTokenIndex])
            startingTokenIndex += 1
            while startingTokenIndex < len(tokenList) and not tokenList[startingTokenIndex].makesSound:
                overflow += str(tokenList[startingTokenIndex])
                startingTokenIndex += 1
    if overflow:
            print overflow + "             *****  <--no rhymes"
#                bestTokenA = bestRhymePairs[0][1]
#                bestTokenB = bestRhymePairs[0][2]
#                printPoetryLine(tokenList[startingTokenIndex:tokenList.index(bestTokenA)], bestTokenA, bestRhymePair[1])
#                printPoetryLine(tokenList[tokenList.index(bestTokenA)+1:tokenList.index(bestTokenB)], bestTokenB, bestRhymePair[2])
#                startingTokenIndex = tokenList.index(bestTokenB)+1

def getBestPossibleLines(rhymePair, startingTokenIndex):
    ''' Takes a pair of tokens and a start point. Grabs the resulting lines from the global tokenList and tries 
    to make them line up through rearrangements and deletions. Returns the best-scoring solution in a list
    containing [score, lines[], tokensWereSwapped, movements[] ]
    '''
    bestTokenA = rhymePair[1]
    bestTokenB = rhymePair[2]
    bestRhymePair = bestRhymes(bestTokenA,bestTokenB)[0]
    endIndex = tokenList.index(bestTokenB)+1
    while(not tokenList[endIndex].isStop):#find last token that might be involved in within-sentence movements. I.e. the end of the second token's sentence.
        endIndex += 1
    window = tokenList[startingTokenIndex : endIndex+1]
    aSyl = bestRhymePair[1].numSyllables()
    bSyl = bestRhymePair[2].numSyllables()
    relevantMovements = getRelevantRearrangements(startingTokenIndex, tokenList.index(bestTokenA), tokenList.index(bestTokenB))
    possibleLines = []
    n = len(relevantMovements)
    if n==0:
        n=1#force the code below to run even if do-nothing is the only option
    if n <= 14:#try all combinations of movements. This is O(2^n) so we need to limit n
        maxCombo = 2 ** n
        for combo in range(maxCombo):#for each possible True/False sequence, apply that combination of movements. 
                                    #Use the binary representation of the "combo" integer to enumerate all possible combinations. 
                                    #Note that the first will be all 0's, which covers the do-nothing option
            perturbedWindow = list(window)
            listOfMovements = []
            movementPenalty = 1.0
            for index, movement in enumerate(relevantMovements):
                if ((combo >> index)&0x01) > 0:
                    listOfMovements.append(movement)
                    movement.applyToTokenList(perturbedWindow, False)#apply the movement and get a new, post-movement sequence of tokens
                    movementPenalty *= movement.penaltyFactor
            score_lines_swap = getEqualizedLinesGivenWindow(perturbedWindow, bestTokenA, bestTokenB, aSyl, bSyl)
            if score_lines_swap:
                score_lines_swap[0] = score_lines_swap[0] * movementPenalty
                score_lines_swap.append(listOfMovements)
                score_lines_swap_movements = score_lines_swap #just documenting what it holds
                possibleLines.append(score_lines_swap_movements)
    else:#if n is too large, just try to apply each movement individually
        relevantMovements.append(MovementList(bestTokenA,1.0))#include a do-nothing movement
        for mList in relevantMovements:#try applying each movement individually
            perturbedWindow = list(window)
            mList.applyToTokenList(perturbedWindow, False)#apply the movement and get a new, post-movement sequence of tokens
            score_lines_swap = getEqualizedLinesGivenWindow(perturbedWindow, bestTokenA, bestTokenB, aSyl, bSyl)
            if score_lines_swap:
                score_lines_swap[0] = score_lines_swap[0] * mList.penaltyFactor
                score_lines_swap.append([mList])
                score_lines_swap_movements = score_lines_swap #documenting what it holds
                possibleLines.append(score_lines_swap_movements)
    if not possibleLines:
        return None
    else:
        possibleLines.sort()
        return possibleLines[-1]#highest score

lastTokenPrintedWasAPeriod = True

def printPoetryLine(listOfTokens, lastToken, synonymOfLastTokenToUse):
    '''Takes a list of RhymeTokens and prints them, but substitutes synonymOfLastTokenToUse for the last one. 
        Continues printing any punctuation after the last token.
        Also does some formatting.'''
    global lastTokenPrintedWasAPeriod
    #print str(listOfTokens) + str(synonymOfLastTokenToUse)
    text = ""
    for token in listOfTokens:#print all tokens in the line, fixing the capitalization if necessary
        if lastTokenPrintedWasAPeriod:
            text += " " + str(token).capitalize()
        else:
            if token.isStart:#it was the first word in a sentence but it no longer is.
                text += " " + str(token).lower()
            else:#standard situation; leave capitalization as is.
                text += " " + str(token)
        lastTokenPrintedWasAPeriod = token.isStop
#        if token.phraseEnd:
#            text += "*"
    #print the last token, substituting in the rhyming word
    if lastTokenPrintedWasAPeriod:
        text += " " + str(synonymOfLastTokenToUse).capitalize()
    else:
        if lastToken.isStart:#it was the first word in a sentence but it no longer is.
            text += " " + str(synonymOfLastTokenToUse).lower()
        else:#standard situation; leave capitalization as is.
            text += " " + str(synonymOfLastTokenToUse)
#    if lastToken.phraseEnd:
#            text += "*"
    lastTokenPrintedWasAPeriod = lastToken.isStop
    ix = tokenList.index(lastToken)+1
    while ix < len(tokenList) and not tokenList[ix].makesSound: #also print trailing punctuation marks
        text += str(tokenList[ix])
        lastTokenPrintedWasAPeriod |= tokenList[ix].isStop
        ix += 1
    start = 0
    punctuation = {" ",",",";",".","?","!"}
    while text[start] in punctuation:
        start+=1
    text = text[start:] #remove leading punctuation marks. These belonged to the previous line and were hopefully printed there.
    text = text.replace("  ", " ")#remove leftovers from deleted phrases
    text = text.replace(" .", ".")
    text = text.replace(" ,", ",")
    text = text.replace(" ;", ";")
    text = text.replace(" :", ":")
    text = text.replace(" -", "-")
    text = text.replace(",,", ",")
    text = text.replace("..", ".")
    text = text.replace(",.", ".")
    text = text.replace("-.", ".")
    
    print text

def getEqualizedLinesGivenWindow(perturbedWindow, bestTokenA, bestTokenB, aSyl, bSyl):
    '''Takes a window (list) of RhymeTokens and the desired rhyming tokens, and tries to equalize the resulting lines by deleting non-essential words. 
        If it succeeds, returns [deletionPenalty, listHoldingEqualizedLines, movementSwappedAAndB]
    '''
    aNewIndex = perturbedWindow.index(bestTokenA)
    bNewIndex = perturbedWindow.index(bestTokenB)
    movementSwappedAAndB = False
    if bNewIndex < aNewIndex:
        movementSwappedAAndB = True
        aNewIndex, bNewIndex = bNewIndex, aNewIndex #fancy Python swap
        aSyl, bSyl = bSyl, aSyl

    lineA = perturbedWindow[0 : aNewIndex]#notation has the form [start:(one past end)]
    lineB = perturbedWindow[aNewIndex+1 : bNewIndex]
    (deletionPenalty, equalizedLines) = equalizeLines(lineA, lineB, aSyl-bSyl, MIN_SYLLABLES_PER_LINE-aSyl, MAX_SYLLABLES_PER_LINE-aSyl, IDEAL_SYLLABLES_PER_LINE-aSyl)
    if equalizedLines:
        return [deletionPenalty, equalizedLines, movementSwappedAAndB]
    else:
        return None

def getRelevantSentences(tokenList, start, endPlusOne):
    #returns all Sentences that have words in the given range in the given list
    sentences = set()
    for i in range(start, endPlusOne):
        sentences.add(tokenList[i].mySentence)
    return list(sentences)

def getRelevantRearrangements(startOfLineA, endOfLineA, endOfLineB):
    #returns all MovementLists that would change the line lengths of the given lines. Might be empty.
    relevantMLists = [] #start empty
    relevantSentences = getRelevantSentences(tokenList, startOfLineA, endOfLineB+1)
    for s in relevantSentences:
        for mList in s.movementLists:
            l = mList.leftMostIndexAffected()
            r = mList.rightMostIndexAffected()
            if (l>=startOfLineA and l<endOfLineB and #lies in the range of interest 
                (not r<endOfLineA) and #isn't just a pointless movement within line A (such a movement doesn't affect line length and so is not of interest)
                (not (r<endOfLineB and l>=endOfLineA))):#isn't just a pointless movement within line B
                relevantMLists.append(mList)
    return relevantMLists
    


DELETABLE_TAGS = {#if these are the head of a subtree, that tree can be deleted without changing meaning. The associated score penalty multiplier will be applied if it is deleted. So a higher penalty multiplier => less important and will be removed first
                  'PP': .80, #Preposition or subordinating conjunction
                  #'CC': .85, #the word and, or, ... I had hoped this would completely delete one of the items being joined, but the basic parser seems to make the conjunction its own branch between the branches it joins, so I'm disabling this.
                  'JJ': .91,#Adjective, not comparative or superlative
                  'RB': .93,#Adverb, not comparative or superlative
                  'CD': .91,#Cardinal number
                  'UH': .97,#Interjection
                  'ROOT': .85,#Whole sentence. NOTE: DeletionList search implementation currently depends on this being here. If you don't like it, change its multiplier to 0
                  }
UNDELETABLE_WORDS = {"no","not","n't"} #these may be parsed as adjectives/adverbs but should not be removed alone. (They can still be removed as part of a larger phrase though)
MOVABLE_TAGS = {#if these are the head of a subtree, that tree can be moved to the head of the phrase (denoted "S" in pennPrint()) without changing meaning
                'PP': .98}#can move preposition to beginning. E.g. "his eyes shone with a steely glitter" -> "with a steely glitter his eyes shone"

class DeletionPlan: #holds a list of DeletionLists
    def __init__(self, dListsToDelete, numSyllables, penaltyFactor):
        self.dLists = dListsToDelete
        self.numSyllables = numSyllables
        self.penaltyFactor = penaltyFactor
    def __repr__(self):
        return "DeletionPlan: to remove " + str(self.numSyllables) + " total syllables, do this: " + str(self.dLists)
    @staticmethod
    def combine(dPlanA, dPlanB):
        newDLists = list(dPlanA.dLists)#copy to make a new list
        newDLists.extend(dPlanB.dLists)#add the other list to it. NOTE: extend() returns None, so you can't chain the method calls. 
                                        #If you try, Python happily assigns "None" to the variable, leaving a landmine for later code and causing me to spend 2 hours debugging.
        nSyl = dPlanA.numSyllables + dPlanB.numSyllables
        newPenalty = dPlanA.penaltyFactor*dPlanB.penaltyFactor
        newDPlan = DeletionPlan(newDLists, nSyl, newPenalty)
#        print "combined " + str(dPlanA)
#        print "and " + str(dPlanB)
#        print "to make " + str(newDPlan)
#        print "its list is: " + str(newDLists)
        return newDPlan
    def applyToLine(self, line):#deletes tokens from the line, in-place (i.e. changes the passed-in list)
        for dList in self.dLists:
            if dList.rhymeTokensToRemove:
                for token in dList.rhymeTokensToRemove:
                    if token in line:
                        line.remove(token)
                    else:
                        print "ERROR: tried to remove: " + str(token)
                        print "when it wasn't in: " + str(line)
                        print "plan was: " + str(self)
    

class DeletionList:#Holds references to a subtree of RhymeTokens that can be removed as a unit, as necessary for shortening a line of poetry 
    def __init__(self, penaltyForRemoval, rhymeTokensToRemove, relation = "not set"):
        self.relation = relation
        self.rhymeTokensToRemove = rhymeTokensToRemove #A list of child nodes which must be removed if this token is removed
        self.penaltyForRemoval = penaltyForRemoval #The penalty factor for removing this node and its children. Between 1=no penalty and 0=do not remove.
        self._nSyllables = -1
        self.allDListsUnderMe = [] #stores ALL dLists which are under this dList. Not just direct children.
        self.children = [] #dLists which are directly under this one  Are disjoint from each other, but we might have words than none of them do.
        self.longestDeletionPlan = None
        self.longestDeletionPlanNumSyllables = -1
    def __repr__(self):
        return "DeletionList: to remove " + str(self.numSyllables()) + " syllables in a "+self.relation+", delete: " + str(self.rhymeTokensToRemove)
    def numSyllables(self):
        if self._nSyllables < 0:
            self._nSyllables=0
            for t in self.rhymeTokensToRemove:
                self._nSyllables += t.synonyms[0].numSyllables()
        return self._nSyllables
    def removeInvalidChildren(self):#returns true if this list is for an undeletable word. Also removes children who returned true.
        for child in self.children:
            shouldBeRemoved = child.removeInvalidChildren()
            if shouldBeRemoved:
                self.children.remove(child)
                self.allDListsUnderMe.remove(child)
        if not self.rhymeTokensToRemove or (len(self.rhymeTokensToRemove)==1 and str(self.rhymeTokensToRemove[0]) in UNDELETABLE_WORDS):
            return True
        else:
            return False
    
    def getDisjointDListsInSet(self, line, maxSyllables):#returns a list containing the heads of my highest DeletionList subtrees that are entirely in the given set and have less than maxSyllables.
        if self.numSyllables() <= maxSyllables:
            imAllInTheSet = True                #implementation: check if we're all in the set. If we are, return self. If we aren't, recursively have all our children do the same and combine what they return. 
            for token in self.rhymeTokensToRemove:
                if not token in line:
                    imAllInTheSet=False
                    break
            if imAllInTheSet:#we're all inside the line, so we're the largest node possible.
                return [self]
        #if we're not entirely in the line, or we're too big, return a list of our children
        dLists = []
        for child in self.children:
            dLists.extend(child.getDisjointDListsInSet(line, maxSyllables))
        return dLists
    @staticmethod
    def addToPlanDictIfBetter(planDict, maxSyllablesForPlanDict, dPlanToAdd):
        '''If a plan is better than the current one in planDict, and it's below maxSyllablesForPlanDict, we'll add it to planDict. Else, discard it.
        '''
        nSyl = dPlanToAdd.numSyllables
        if nSyl > maxSyllablesForPlanDict or nSyl < 1:
            return
        if not nSyl in planDict or planDict[nSyl].penaltyFactor < dPlanToAdd.penaltyFactor: #dict has no solution for our number of syllables, or its solution is worse than ours
            planDict[nSyl] = dPlanToAdd
        return
        
    
    def getBestDeletionPlans(self, maxSyllables):#returns a dict of the best DeletionPlan we have for deleting each of 1,2,3,...maxSyllables syllables.
        '''Algorithm: At each node, make a dict containing possible DeletionPlans, indexed by the number of syllables they'll remove. Only keep the best
        DeletionPlan for each number of syllables. When combining child dicts, keep a running best plan, and for each child, try replacing each entry, 
        as well as all combinations of previous entries and new entries. In the end, we'll end up with the best way to remove any number of syllables below maxSyllables. '''
        #if maxSyllables > self.longestDeletionPlanNumSyllables:#see if we can use the cached version. BROKEN - need to think about why this fails at the boundary of a line.
        if self.children:
            existingBestPlans = self.children[0].getBestDeletionPlans(maxSyllables) #dictionary holding best DeletionPlans so far. Indexed by number of syllables that the plan will delete.
            #now add all combinations of child deletion plans. Note that child deletion plans can replace OR combine with those of other children.
            
            for i in range(1, len(self.children)):
                updatedPlan = existingBestPlans.copy() #will hold a new plan with the best combinations of the existing plans and this child's
                childsPlan = self.children[i].getBestDeletionPlans(maxSyllables)
                for nSyl in childsPlan:#for each plan in the dictionary
                    DeletionList.addToPlanDictIfBetter(updatedPlan, maxSyllables, childsPlan[nSyl])#try replacing the current plan for this number of syllables. I.e. combine with the "0 deletions" option from the other plan.
                    for oldPlanNSyl in existingBestPlans:#try combination with all previous plans
                        if oldPlanNSyl < maxSyllables-nSyl:#only look at addition if it won't put us over the limit (this also gets checked by addToPlan, but pruning helps)
                            combination = DeletionPlan.combine(childsPlan[nSyl], existingBestPlans[oldPlanNSyl])
                            DeletionList.addToPlanDictIfBetter(updatedPlan, maxSyllables, combination)
                existingBestPlans = updatedPlan
        else:
            existingBestPlans = dict()
        if self.numSyllables() > 0:
            dPlanForJustMe = DeletionPlan([self], self.numSyllables(), self.penaltyForRemoval)
            DeletionList.addToPlanDictIfBetter(existingBestPlans, maxSyllables, dPlanForJustMe)
        self.longestDeletionPlan = existingBestPlans
        self.longestDeletionPlanNumSyllables = maxSyllables
        return self.longestDeletionPlan
        

class MovementList:#hold a list of RhymeTokens which can be moved to a certain destination without affecting sentence meaning.
    def __init__(self, rhymeTokenToInsertBefore, penaltyFactor):
        self.rhymeTokensToMove = [] #these tokens
        self.rhymeTokenToInsertBefore = rhymeTokenToInsertBefore #can be moved to go right before this token.
        self.penaltyFactor = penaltyFactor
        #print "made new MList to go before " + str(self.rhymeTokenToInsertBefore)
    def __repr__(self):
        return "MovementList: move " + str(self.rhymeTokensToMove) + " to just before " + str(self.rhymeTokenToInsertBefore)
    def leftMostIndexAffected(self):#tokens which are left of this index in the global list won't be affected by the move. Used to make sure we don't break lines we've already written. Also to check if applying this will have any effect on a subsection of interest.
        return self.rhymeTokenToInsertBefore.index
    def rightMostIndexAffected(self):
        return self.rhymeTokensToMove[-1].index
    def applyToTokenList(self, line, updateTokenIndices):#Does the movement to the given line. The line may or may not be the global list. Only set updateTokenIndices to True if it is
        for tokenToMove in self.rhymeTokensToMove:#this is inefficient but whatever.
            line.remove(tokenToMove) #we're working with RhymeToken objects, not strings, so each token object is unique, so we can do this with no worries.
        indexToInsert = line.index(self.rhymeTokenToInsertBefore)
        line[indexToInsert:indexToInsert] = self.rhymeTokensToMove #insert these items into that zero-width slice
        if updateTokenIndices:#we're on the global list. Update the index tracking of the rhyme tokens we've affected
            for idx, token in enumerate(line):
                token.index = idx





class SentenceTree:
    def __init__(self):
        self.parseTree = None
        self.deletionLists = None
        self.movementLists = None
    def fill(self, parseTree, deletionLists, movementLists):
        self.parseTree = parseTree
        self.deletionLists = deletionLists
        self.movementLists = movementLists

def buildTokenList(sentenceStrings, sentenceTrees):
    global currentSentencesDLists, currentSentencesMLists, currentRhymeTokens #tells python that we're talking about the global one, not a local variable
    currentRhymeTokens = []
    for sentence in sentenceStrings:
        sTree = SentenceTree()
        parseTree = lp.apply(sentence); #now we have a parseTree tree and can do what we want with it
        currentSentencesDLists = [] #TODO: track what sentence each RhymeToken is in
        currentSentencesMLists = []
        startIndex = len(currentRhymeTokens)
        exploreTree(parseTree,currentRhymeTokens) #Recursively builds the list of currentRhymeTokens and deletionLists
        currentRhymeTokens.append(RhymeToken("\n", len(currentRhymeTokens)))#indicates end of a sentence
        endIndex = len(currentRhymeTokens)
        for i in range(startIndex, endIndex):
            currentRhymeTokens[i].mySentence = sTree#give all the new tokens a reference to their Sentence object
        currentRhymeTokens[startIndex].isStart = True
        for dList in currentSentencesDLists:#as kind of a hack, I go back after the fact and remove dLists that are bad for whatever reason
            shouldRemove = dList.removeInvalidChildren()
            if shouldRemove:
                currentSentencesDLists.remove(dList)
        sTree.fill(parseTree, currentSentencesDLists, currentSentencesMLists)
        sentenceTrees.append(sTree)
        #parseTree.pennPrint()
#        for mList in currentSentencesMLists:
#            print str(mList)
    return currentRhymeTokens







#linear information used by exploreTree:
def markLastTokenAsAPhraseEnd():
    currentRhymeTokens[-1].phraseEnd = True;
    
lastPhraseHead = None #point to the head word of the "S" phrase currently being explored. "None" indicates we should set it to the first word we see.
currentSentencesDLists = None
currentSentencesMLists = None
currentRhymeTokens = None

def exploreTree(thisNode, rhymeTokens, parentsPOS=None, dListsWeAreIn=None, mListsWeAreIn=None):
    ''''Recursively processes a tree, putting the words into rhymeTokens and putting DeletionLists into allDeletionLists
    inputs:
        thisNode - The current node
        rhymeTokens - the list of words, in order. Will add this word to it.
        allDeletionLists - newly-discovered DeletionLists should be added to this
        parentsPOS - The parent of a word is always the part of speech label for that word. So we need to pass that to the child.
        dListsWeAreIn - contains references to DeletionLists that we're in. We must add the current node to all of them if it's a word. 
                        If we are the head of a new deletion list, we need to create a new one of these for our children. 
                        The new dListsWeAreIn will contain all upper DeletionLists plus the new DeletionList.
    '''
    global lastPhraseHead, currentSentencesDLists, currentSentencesMLists #tells Python that we mean to write to the global variable when we write to this
    if dListsWeAreIn is None:
        dListsWeAreIn=[]
    if mListsWeAreIn is None:
        mListsWeAreIn=[]
    
    #print "value: "+thisNode.label().value()
    label = thisNode.label()
    if label.word():#node holds a word.
        token = RhymeToken(label.word(), len(rhymeTokens), parentsPOS)
        token.dListsImIn = dListsWeAreIn
        rhymeTokens.append(token) #add it to the original-order word list
        if lastPhraseHead is None:#we're the first word in a sentence or "S" phrase
            lastPhraseHead = token
        for l in dListsWeAreIn: #if we're in any deletable subtrees, add ourselves to their DeletionLists
            l.rhymeTokensToRemove.append(token)
        for l in mListsWeAreIn: #if we're in any deletable subtrees, add ourselves to their DeletionLists
            l.rhymeTokensToMove.append(token)
        #print "word: "+label.word() + " tag: " + parentsPOS
    else:#node is a dependency relation
        relation = label.value()
        if relation in DELETABLE_TAGS: #we're the head of a deletable subtree
            #print "deletable "+relation+":"
            newDList = DeletionList(DELETABLE_TAGS[relation], [], relation) #our children will be added to this DeletionList
            for l in dListsWeAreIn: #if we're in any deletable subtrees, add our dList as a child of their DeletionLists
                l.allDListsUnderMe.append(newDList)
            if dListsWeAreIn:
                dListsWeAreIn[-1].children.append(newDList)#register this list as a direct child of the most recent dList 
            dListsWeAreIn = list(dListsWeAreIn) #make an identical copy of our parent's list of DeletionLists, and change to point to it
            dListsWeAreIn.append(newDList) #add the new DeletionList to the new copy
            currentSentencesDLists.append(newDList) #also add it to the global list of DeletionLists
        if relation in MOVABLE_TAGS and not lastPhraseHead is None:
            newMList = MovementList(lastPhraseHead, MOVABLE_TAGS[relation])
            mListsWeAreIn = list(mListsWeAreIn) #make an identical copy of our parent's list of DeletionLists, and change to point to it
            mListsWeAreIn.append(newMList) #add the new DeletionList to the new copy
            currentSentencesMLists.append(newMList)
        if relation == "S":
            lastPhraseHead = None
        parentsPOS = relation
    for tree in thisNode.children():
        exploreTree(tree, rhymeTokens, parentsPOS, dListsWeAreIn, mListsWeAreIn)
    if len(thisNode.children()) > 1:#When the above returns, our rightmost child has just been added. if we were the head of a multi-word phrase, mark our rightmost child as a phrase end.
        markLastTokenAsAPhraseEnd();


def equalizeLines(lineA, lineB, bLongerThanAByThis, minA, maxA, idealA):#takes two lists of words and tries to delete some of them to equalize the line length in syllables
    lineTokensAsASet = [set(lineA), set(lineB)]
    lines = [lineA, lineB]
    lineLength = [0,0]
    sentenceDListHeads = [ set(), set() ] #holds sentence heads that might be relevant to each line
    bestPlansForDeletion = [dict(), dict()] #a dict of {syllablesToRemove: best DeletionPlan that does it}
    bestPlansForLength = [dict(), dict()] #a dict of {lineLengthInSyllables: best DeletionPlan that does it}
    for lineNum in range(len(lines)):
        for token in lines[lineNum]:
            if token.dListsImIn:
                sentenceDListHeads[lineNum].add(token.dListsImIn[0])#the first dList is the sentence head. We'll be able to follow the tree-like structure downward from it.
            lineLength[lineNum] += token.synonyms[0].numSyllables()
    
    if lineLength[0] < minA or lineLength[1] < minA+bLongerThanAByThis:
        return (0, False) #can't solve our problems by deletion
    
    lineLengthOffset = [0,bLongerThanAByThis] #line should be longer than the baseline, by this amount (accounts for immutable end words)
    agreeableLengths = set(range(minA, maxA))#will end up with the intersection of possible line lengths for all lines
    
    for lineNum in range(len(lines)):
        maxSyllablesToRemove = lineLength[lineNum] - (min(agreeableLengths) + lineLengthOffset[lineNum])
        if maxSyllablesToRemove < 0:#this line can't be made to work
            return (0, False)
        umbrellaDList = DeletionList(0, []);#temporary holder for the unique tree heads whose words are entirely inside the line, and therefore can be used.
        for dList in sentenceDListHeads[lineNum]:
            umbrellaDList.children.extend(dList.getDisjointDListsInSet(lineTokensAsASet[lineNum], maxSyllablesToRemove))
        ''''now, to find the combination of removals that will give us the right length, we have a "0-1 knapsack problem." 
        It takes O(dLists * maxSyllablesToRemove^2) time.  '''
        bestPlansForDeletion[lineNum] = umbrellaDList.getBestDeletionPlans(maxSyllablesToRemove) #get a dict of {syllablesToRemove: best DeletionPlan that does it}
        bestPlansForLength[lineNum] = {lineLength[lineNum] - lineLengthOffset[lineNum]- 0 : DeletionPlan([], 0, 1)}
        myPossibleLengths = {lineLength[lineNum] - lineLengthOffset[lineNum]- 0 }#Start with the fact that  we can remove 0 syllables. Also convert our length into terms of the baseline
        for nSylToRemove in bestPlansForDeletion[lineNum]: #for each of our DeletionPlans, calculate the line length it can give us, including slop, and record it in the dict bestPlansForLength.
            for slop in range(-SLOP_IN_LENGTH_MATCHING, SLOP_IN_LENGTH_MATCHING+1):#range goes to end-1
                possibleLength = lineLength[lineNum] - lineLengthOffset[lineNum]- nSylToRemove+slop #convert our length into terms of the baseline
                if not possibleLength in bestPlansForLength[lineNum]:# or bestPlansForLength[lineNum][possibleLength].penaltyFactor < bestPlansForDeletion[lineNum][nSylToRemove].penaltyFactor
                    myPossibleLengths.add(possibleLength)
                    bestPlansForLength[lineNum][possibleLength] = bestPlansForDeletion[lineNum][nSylToRemove]
        
        agreeableLengths = set.intersection(agreeableLengths, myPossibleLengths)    
        if not agreeableLengths: #not possible to match the lines; go ahead and abort
            return (0, False)
    
    #now we have a list of possible matching lengths. Choose the one closest to our ideal.
    bestLength = 0
    bestPenalty = -1
    for l in agreeableLengths:
        penaltyToGetLength = 1.0
        for lineNum in range(len(lines)):
            penaltyToGetLength *= bestPlansForLength[lineNum][l].penaltyFactor
        penaltyToGetLength *= (1-PENALTY_PER_DIFFERENCE_FROM_IDEAL*abs(l-idealA))
        if penaltyToGetLength > bestPenalty:
            bestLength = l
            bestPenalty = penaltyToGetLength
    #print "finalLength is "+str(finalLength) +", chosen from " + str(agreeableLengths)
    #score = 1.0
    for lineNum in range(len(lines)):
        bestPlansForLength[lineNum][bestLength].applyToLine( lines[lineNum] )
        #score *= bestPlansForLength[lineNum][bestLength].penaltyFactor
    return (bestPenalty, lines)


print "reading file..."
f=open("./excerpt.txt", "r")
wholeFileAsString = f.read();
wholeFileAsString = wholeFileAsString.replace("\n"," ")#more preformatting may be necessary depending on the text
wholeFileAsString = wholeFileAsString.replace("-"," - ")#more preformatting may be necessary depending on the text
wholeFileAsString = wholeFileAsString.replace("  "," ")

totalStartTime = datetime.datetime.now()
sentenceStrings = tokenize.sent_tokenize(wholeFileAsString)
print "parsing sentences..."
sentenceTrees = []
tokenList = buildTokenList(sentenceStrings, sentenceTrees) #holds RhymeTokens in the original word sequence. TODO: do this for just the next few sentences, on demand.

    
#now sentenceLists is a list of lists of RhymeTokens

''' The plan: search a range of words for the end of line 1. 
Skip ahead and search a range for the end of line 2. Choose the best match. Repeat for 3&4, 5&6, etc
'''

MIN_SYLLABLES_PER_LINE = 5 #min length of an output rhyming line
MAX_SYLLABLES_PER_LINE = 13 #max length of an output rhyming line
IDEAL_SYLLABLES_PER_LINE = 9 #If there are multiple possible lines in the above range, we'll choose the one closest to this length.
PENALTY_PER_DIFFERENCE_FROM_IDEAL = .02 #a penalty multiplier of (1-this*difference) gets applied if a line length solution differs from IDEAL_SYLLABLES_PER_LINE
MAX_SYLLABLES_TO_SEARCH_OVER = MAX_SYLLABLES_PER_LINE*7 #for performance reasons, we'll only look forward over this many syllables of the original text, when trying to make a rhyme 
SLOP_IN_LENGTH_MATCHING = 1 #max difference between two lines will be twice this.
BONUS_FACTOR_FOR_RHYME_ON_PHRASE_END = 1.4 #ends of phrases tend to be natural pauses, so rhymes sound better there. This bonus multiplier will be applied to rhymes that occur at phrase ends.


print "searching for rhymes and aligning rhyming lines...\n"

searchStartTime = datetime.datetime.now()
#printBestSingleRhymeForTesting(tokenList)
#printBestRhymesEqualized(tokenList)
#printSearchedRhymesEqualized(tokenList)
printSearchedRhymesRearranged(tokenList)

endTime = datetime.datetime.now()
print "\nelapsed after startup: " + str(endTime-totalStartTime)
print "elapsed during searches: " + str(endTime-searchStartTime)

'''
userText0 = raw_input("Type a word: ")
userText1 = raw_input("Type another word: ")
#tokens = nltk.word_tokenize(userText)
#tagged = nltk.pos_tag(tokens)
#print tagged
#print wordnet.lin_similarity(synset1, synset2, ic, verbose)
#or:
#synset1.path_similarity(synset2)
#also:
#nltk.classify can do naive bayes and other classifiers


synonyms0 = addPhonemes( getSynonyms(userText0,2) ); #get an array containing a bunch of [word, phonemes] tuples
synonyms1 = addPhonemes( getSynonyms(userText1,2) );
rhymePairs = []
for wordTuple0 in synonyms0:
    for wordTuple1 in synonyms1:
        if(wordTuple0[0].rstrip()[-1] != wordTuple1[0].rstrip()[-1]): 
            score = poetry.rhyme_score_from_phonemes(wordTuple0[1], wordTuple1[1])
            if(score >= 1):
                rhymePairs.append( [ score,wordTuple0[0],wordTuple1[0] ] )
                #print "Rhyme: " + str(wordTuple0[0]) + ", " + str(wordTuple1[0])+", "+str(score)
rhymePairs = sorted(rhymePairs, key=lambda element: -element[0]) #sort, highest first


for pair in rhymePairs:
    print pair
print "end of rhymes."
'''

jpype.shutdownJVM() #call at the end of the script




#notes:
#use wordnet.Morphy to switch between verb/noun forms
#stopwords has lists of stopwords